@@ -1,15 +1,12 @@
-import functools
 import logging
 import math
 import os
 import time
-import tqdm
 from datetime import datetime
 from pathlib import Path
 
-import numpy as np
 import torch
-
+import tqdm
 from IPython.core.display import HTML, display
 
 from . import drawers
@@ -63,6 +60,8 @@ class Miner(object):
         plugins (list, optional):
             Defaults to []. This is actually a collection of `hooks`, do not set
             hooks and plugins the same time.
+        forward (function):
+            custom forward function.
     """
 
     def __init__(
@@ -71,7 +70,8 @@ class Miner(object):
             resume=True, eval_stride=1, persist_stride=1, gpu=True,
             drawer='matplotlib', hooks={}, max_epochs=None, statable={},
             logging_format=None, trival=False, in_notebook=False, plugins=[],
-            logger=None, sheet=None, accumulated_iter=1, ignore_optimizer_resume=False):
+            logger=None, sheet=None, accumulated_iter=1, ignore_optimizer_resume=False,
+            forward=None):
         self.alchemistic_directory = alchemistic_directory
         self.code = code
         if trival:
@@ -109,6 +109,7 @@ class Miner(object):
         self.hook_funcs = hooks
         self.max_epochs = max_epochs
         self.trival = trival
+        self.forward_fn = forward
 
         self.sheet = sheet
         if self.sheet:
@@ -249,8 +250,8 @@ class Miner(object):
 
             try:
                 self.model.load_state_dict(checkpoint['state_dict'], strict=True)
-            except:
-                msg = ('load checkpoint failed, the state in the '
+            except Exception as e:
+                msg = (f'load checkpoint failed with {e}, the state in the '
                        'checkpoint is not matched with the model, '
                        'try to reload checkpoint with unstrict mode')
                 self.logger.warning(msg)
@@ -260,9 +261,9 @@ class Miner(object):
             if 'optimizer' in checkpoint and not self.ignore_optimizer_resume:
                 try:
                     self.optimizer.load_state_dict(checkpoint['optimizer'])
-                except:
-                    msg = ('load optimizer state failed, will skip this error and continue, '
-                            'stop the process if it is not expected')
+                except Exception as e:
+                    msg = (f'load optimizer state failed with {e}, will skip this error and continue, '
+                           'stop the process if it is not expected')
                     self.logger.warning(msg)
                     self.notebook_output(msg)
 
@@ -290,7 +291,6 @@ class Miner(object):
                 model = torch.nn.DataParallel(model, devices)
         return model
 
-
     def notify(self, message, _type='info'):
         getattr(self.logger, _type)(message)
         self.notebook_output(message, _type)
@@ -409,11 +409,7 @@ class Miner(object):
             index=index,
             total_iters=train_iters,
             iteration=self.current_train_iteration)
-        predict = self.model(data[0].to(self.devices))
-        # for the last batch, loss is not supposed divide by self.accumulated_iter
-        # just ignored this tiny issue
-        # loss = self.loss_func(predict, *data[1:])
-        loss = self.loss_func(predict, data[1].to(self.devices))
+        _, loss = self._forward(data)
         seperate_loss = loss / self.accumulated_iter
         seperate_loss.backward()
         loss = loss.detach().cpu().item()
@@ -430,6 +426,14 @@ class Miner(object):
         )
         return loss
 
+    def _forward(self, data):
+        if self.forward_fn:
+            return self.forward_fn(self, data)
+        else:
+            predict = self.model(data[0].to(self.devices))
+            loss = self.loss_func(predict, data[1].to(self.devices))
+            return predict, loss
+
     def run_val_iteration(self, index, data, val_iters):
         self.status = 'val'
         self.current_val_iteration += 1
@@ -440,9 +444,7 @@ class Miner(object):
             total_iters=val_iters,
             iteration=self.current_val_iteration
         )
-        predict = self.model(data[0].to(self.devices))
-        # loss = self.loss_func(predict, *data[1:])
-        loss = self.loss_func(predict, data[1].to(self.devices))
+        predict, loss = self._forward(data)
         loss = loss.detach().cpu().item()
         self.logger.info('[val {}/{}/{}] loss {}'.format(
             self.current_epoch, index, val_iters, loss))
@@ -555,8 +557,7 @@ class Miner(object):
             return
 
         self.sheet_progress.update(kwargs)
-        progress = \
-f"""
+        progress = f"""
          epoch:  {self.sheet_progress.get('epoch')}
 train progress:  {self.sheet_progress.get('train_percentage')}
   val progress:  {self.sheet_progress.get('val_percentage')}
